/* Name: servo.c
 * Project: Control two servo motors over USB
 * Author: Pavel Petrovic, robotics.sk
 * Creation Date: 2009-06-21
 * Copyright: (c) 2009 by Robotika.SK
 * License: GNU GPL v2 (see License.txt),
            This program is derived from PowerSwitch.c of Christian Starkjohann
            and OBJECTIVE DEVELOPMENT Software GmbH driver
 */

/*
General Description:
This program controls two servo motors using USB port from the command line.
It must be linked with libusb. http://libusb.sourceforge.net/.
It was tested on Win32 platform, but why shouldn't it work on others.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */

#include <ncurses.h>

/* Use obdev's generic shared VID/PID pair and follow the rules outlined
 * in firmware/usbdrv/USBID-License.txt.
 */
#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */

/* These are the vendor specific SETUP commands implemented by our USB device */
#define PSCMD_ECHO   0
#define PSCMD_GET    1
#define PSCMD_POS    2
#define PSCMD_ADD    3
#define PSCMD_SETA   4
#define PSCMD_SETB   5

static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
char    buffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
        return rval;
    if(buffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)buffer[0] < rval)
        rval = (unsigned char)buffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = buffer[2 * i];
        if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}


/* PowerSwitch uses the free shared default VID/PID. If you want to see an
 * example device lookup where an individually reserved PID is used, see our
 * RemoteSensor reference implementation.
 */

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL){  /* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }else{
                    errorCode = USB_ERROR_NOTFOUND;
                    //fprintf(stderr, "seen device from vendor ->%s<-\n", string);
                    if(strcmp(string, vendorName) == 0){
                        len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                        if(len < 0){
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
                        }else{
                            errorCode = USB_ERROR_NOTFOUND;
                            //fprintf(stderr, "seen product ->%s<-\n", string);
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}

void usage()
{
	printf("\nServo <cmd>\n\n");
	printf("  recognized commands:\n");
	printf("    test - test usb device\n");
	printf("    status - verify usb device status\n");
	printf("    pos - get motor position\n");
	printf("    + A B - modify motor positions\n");
	printf("    !a new - set motor A position\n");
	printf("    !b new - set motor B position\n");
	printf("    manual - manual control: a,z,o,p,q\n\n");
}

int main(int argc, char **argv)
{
usb_dev_handle      *handle = NULL;
unsigned char       buffer[8];
int                 nBytes;

    if(argc < 2){
        usage();
        exit(1);
    }
    usb_init();
    if(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "robotics.sk", USBDEV_SHARED_PRODUCT, "SapienIRCtl") != 0){
        fprintf(stderr, "Could not find USB device \"SapienIRCtl\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
        exit(1);
    }
/* We have searched all devices on all busses for our USB device above. Now
 * try to open it and perform the vendor specific control operations for the
 * function requested by the user.
 */
    if(strcmp(argv[1], "test") == 0){
        int i, v, r;
/* The test consists of writing 1000 random numbers to the device and checking
 * the echo. This should discover systematic bit errors (e.g. in bit stuffing).
 */
        for(i=0;i<1000;i++){
            v = rand() & 0xffff;
            nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
            if(nBytes < 2){
                if(nBytes < 0)
                    fprintf(stderr, "USB error: %s\n", usb_strerror());
                fprintf(stderr, "only %d bytes received in iteration %d\n", nBytes, i);
                fprintf(stderr, "value sent = 0x%x\n", v);
                exit(1);
            }
            r = buffer[0] | (buffer[1] << 8);
            if(r != v){
                fprintf(stderr, "data error: received 0x%x instead of 0x%x in iteration %d\n", r, v, i);
                exit(1);
            }
        }
        printf("test succeeded\n");
    }else if(strcmp(argv[1], "status") == 0){
        nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_GET, 0, 0, (char *)buffer, sizeof(buffer), 5000);
        if(nBytes < 2){
            if(nBytes < 0)
                fprintf(stderr, "USB error: %s\n", usb_strerror());
            fprintf(stderr, "only %d bytes status received\n", nBytes);
            exit(1);
        }
        printf("Status: %d,%d\n", buffer[0], buffer[1]);
    }else if (strcmp(argv[1], "pos") == 0){
		nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_POS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
        if(nBytes < 4){
            if(nBytes < 0)
                fprintf(stderr, "USB error: %s\n", usb_strerror());
            fprintf(stderr, "only %d bytes position received\n", nBytes);
            exit(1);
        }
        printf("%d %d\n", (buffer[0] << 8) + buffer[1], (buffer[2] << 8) + buffer[3]);
    }else if (strcmp(argv[1], "+") == 0){
		if (argc < 4)
		{
			printf("missing arguments (... + dA dB)\n");
			exit(1);
		}
		int da, db;
		int a;
		sscanf(argv[2], "%d", &da);
		sscanf(argv[3], "%d", &db);
		a = ((db & 255) << 8) + (da & 255);
		nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_ADD, a, 0, (char *)buffer, sizeof(buffer), 5000);
    }else if (strcmp(argv[1], "!a") == 0){
		if (argc < 3)
		{
			printf("missing arguments (... !a <A>)\n");
			exit(1);
		}
		int a;
		sscanf(argv[2], "%d", &a);
		nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETA, a, 0, (char *)buffer, sizeof(buffer), 5000);
    }else if (strcmp(argv[1], "!b") == 0){
		if (argc < 3)
		{
			printf("missing arguments (... !b <B>)\n");
			exit(1);
		}
		int b;
		sscanf(argv[2], "%d", &b);
		nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETB, b, 0, (char *)buffer, sizeof(buffer), 5000);
    }else if (strcmp(argv[1], "manual") == 0){
		char c;
		int fl = fcntl(0, F_GETFL);
		fcntl(0, F_SETFL, fl | O_NONBLOCK);
		nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_POS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
        int a = (buffer[0] << 8) + buffer[1];
        int b = (buffer[2] << 8) + buffer[3];
        initscr();
        cbreak();
        noecho();
        nonl();
        intrflush(stdscr, FALSE);
        keypad(stdscr, TRUE);

		do {

			c = getch();
			switch(c)
		    {
				case 'a': a += 10;
				          nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETA, a, 0, (char *)buffer, sizeof(buffer), 5000);
                          break;
				case 'z': a -= 10;
				          nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETA, a, 0, (char *)buffer, sizeof(buffer), 5000);
                          break;
				case 'p': b += 10;
				          nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETB, b, 0, (char *)buffer, sizeof(buffer), 5000);
                          break;
				case 'o': b -= 10;
				          nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SETB, b, 0, (char *)buffer, sizeof(buffer), 5000);
                          break;
		    }
		} while (c != 'q');
		endwin();
    }else{
            usage();
            exit(1);
    }
    usb_close(handle);
    return 0;
}
